/*
    Copyright (c) 2021 PIXILAB Technologies AB, Sweden (http://pixilab.se). All Rights Reserved.
    Created 2021 by Hugo Beck-Friis.
*/

/*
    For this script to work, the Google Calendar Id needs to be set as a parameter ("calendarId") on
	the spot showing the calendar. This calendar id can be found in the Google Calendar settings for
	the calendar you want to show.

    This script uses Google Service Accounts to access private calendars. These can be created in the
    Google Cloud Console. When you have created the account, save the credentials, and invite the
    service account e-mail address to the calendar. If you don't need the calendar to be private,
    having the calendar set to public is an easier and more efficient alternative. In that case,
    please use the script "GoogleCalendarFeed.ts", provided in the application notes for this project.

    Before starting, pelase make sure that the correct service account credentials are set in the
    "credentials" constant. These are obtained when creating the service account in the Google Cloud
    Console.

    In addition to this script, a way to automatically reload the spot is recommended. This can be
	done using multiple methods, e.g. tasks. If you are using a task, make sure to mark "reloadBrowser"
    as true.

    Some libraries used in this code have been slightly modified to work with Blocks (as well as
    removing some unused code). The original code can be found at https://github.com/travist/jsencrypt.
    All licenses are included in the download.

    For more information on authenticating Service Accounts, please see
    https://developers.google.com/identity/protocols/oauth2/service-account#httprest
*/


import {field, id} from "system_lib/Metadata";
import * as feed from "system_lib/Feed";
import {ListData} from "system_lib/Feed";
import {SimpleHTTP} from "system/SimpleHTTP";

// Libraries needed for signing the JWT (JSON Web Token) used when obtaining the access_token from the OAuth authentification of the service account.
import {JSEncrypt} from "lib/JSEncrypt";
import {SHA256} from "lib/webtoolkit_SHA256";
import {Base64} from "lib/webtoolkit_base64";

// The scopes needed to access the calendar events. If more functionality is needed, more can be added (seperated by a space).
const scopes : string = 'https://www.googleapis.com/auth/calendar.readonly';

// Credentials used when authenticating the service account
const credentials = {
    YOUR CREDENTIALS HERE    
};


export class ServiceAccountCalendarFeed extends feed.Feed {
    constructor(env: feed.FeedEnv) {
		super(env);
		// Fields to be available through blocks. The number represents the amount of hours from today's midnight.
		// "TodayPast" represents events that have already happened, and "Today" represents events that will happen.
		// If used, "Today"'s timeOffset-value should always be set to 0, and the same value for "TodayPast" should always be -1. Other days should always be (number of days * 24). More days can be added, if needed.
		this.establishFeed(new Cal("TodayPast", -1));
		this.establishFeed(new Cal("Today", 0));
		this.establishFeed(new Cal("Tomorrow", 24));
		this.establishFeed(new Cal("TwoDays", 48));
	}
}

class Cal implements feed.StaticFeed<ListItem, DetailsItem> {
	// Used to store the data from the request in getList, so it can later be reused in getDetails without sending a completely new request.
	private listResponse : EventListItem[] | DetailsSingleItem;


	// The name visible in the block builder
	readonly name: string;
	readonly timeOffset: number;

	constructor(feedName: string, timeOffset:number) {
		this.name = feedName;
		this.timeOffset = timeOffset;
	}

	// Mandatory fields describing what types of list and item data is provided
	readonly listType = ListItem;
	readonly itemType = DetailsItem;

	// Mandatory method used to get a list of items.
	getList(spec:feed.FeedListSpec): Promise<ListData<ListItem>> {
        // The first thing we need is an access_token for the service account, which is used in place of the API-key from the public calendar example.
        // This access_token is generated by Google, and is obtained through a POST-request with a signed JTW.
        // A JWT is a JSON Web Token, and consists of three parts, seperated by dots:
        // 1. A Header. When communicating with OAuth, this should always be '{"alg":"RS256","typ":"JWT"}', encoded using Base64, with URL and Filename Safe Alphabet. We use the variable "jwtHeader" for this.
        // 2. A Claim. This is the important part, and contains information about the service account (again encoded with Base64). In our case, this is the variable "jwtClaim".
        // 3. A Signature. The only signing algorithm allowed by the OAuth Authorization Server is "RSA using SHA-256 hashing algorithm", expressed as "RS256" in the "alg"-field in the header.
        //    The signature uses the private key from the service account to sign the other two fields in the following format: "{Base64url encoded header}.{Base64url encoded claim set}". It is lated encoded wuth Base64URL itself.

        // Get the current time (UNIX format). This is used in the JWT claim to set the start- and expiration times for the access_token that is returned.
        let timeNow : number = Math.floor(Date.now() / 1000);

        // JTW Header. When communicating with OAuth, this is always the object used
        let jwtHeader : string = JSON.stringify({"alg":"RS256","typ":"JWT"});

        // JWT Claim.
        let jwtClaim = JSON.stringify({
            "iss": credentials.client_email, // The e-mail address of the service account
            "scope": scopes, // The scopes used when accessing the API. If you need more than calendar.readonly, please change the scopes constant.
            "aud": "https://oauth2.googleapis.com/token", // Target of the assertion. Do not change this.
            "exp": timeNow + 3600, // Expiration time. Set to one hour from the current time (which is the maximum time Google allows).
            "iat": timeNow // Issue time (start)
        });

        // Initiate the JWS signing process by creating a new JSEncrypt instance and setting the private key to the service account's private key
        var sign : JSEncrypt = new JSEncrypt({});
        sign.setPublicKey(credentials.private_key);

        // Create the signature. The library used for Base64-encoding does not have support for Base64URL, so we have to make sure the format is correct.
        var jwtSignature : string | false =
            sign.sign(
                Base64.encode(jwtHeader)
                    .replace(/=+$/, '')
                    .replace(/\//g, '_')
                    .replace(/\+/g, '-') +
                "." +
                Base64.encode(jwtClaim)
                    .replace(/=+$/, '')
                    .replace(/\//g, '_')
                    .replace(/\+/g, '-'),
                SHA256, // Hashing function, from webtoolkit_SHA256
                "sha256" // Specify the hash type
            );

        // Combine the header, claim, and signature. This is the string that will be used in the POST request that obtains the access_token.
        var assertionString : string =
            Base64.encode(jwtHeader)
                .replace(/=+$/, '')
                .replace(/\//g, '_')
                .replace(/\+/g, '-') +
            "." +
            Base64.encode(jwtClaim)
                .replace(/=+$/, '')
                .replace(/\//g, '_')
                .replace(/\+/g, '-') +
            "." +
            jwtSignature.toString()
                .replace(/=+$/, '')
                .replace(/\//g, '_')
                .replace(/\+/g, '-');

		// Get the time, which will later be reset to the beginning of the day.
		// This time will used in the second API request (getting calendar events)
		let timeBase: Date = new Date();

		// Offset the time based on the timeOffset variable, and create a string in a format readable by the Calendar API.
		// The number in the variable represents hours. When the 24h-mark is passed, the hours will reset to 0, and one day will be added.
		// The variable we create will be used to define what start-point we want the request to return data from.
		// "Today" is an exception to this because it starts when "TodayPast" ends (i.e. at the time the feed is loaded)
		if (this.timeOffset != 0) {
			timeBase.setUTCHours(this.timeOffset, 0, 0, 0);
		}

        let timeMinString : string = timeBase.toISOString();

		// Jump forward to midnight from the previous offset and create a new string that is used to tell the API when to stop loading events.
		// "TodayPast" is an exception to this because it ends when "Today" starts (i.e. at the time the feed is loaded)
		if (this.timeOffset != -1) {
			timeBase.setUTCHours(24, 0, 0, 0);
		} else {
			timeBase = new Date();
		}

		let timeMaxString : string = timeBase.toISOString();

		// Get the access_token via a POST-requst, and then use it to make a new request getting the calendar events
		return SimpleHTTP
            .newRequest(
                    "https://oauth2.googleapis.com/token",
                    { interpretResponse: true }
                )
                .setTimeout(25) // Prevent the request from taking too long
                .post( // The data to POST, in this case the JWT. The grant_type should always be "urn:ietf:params:oauth:grant-type:jwt-bearer", in a URL-friendly format.
                    "grant_type=urn%3Aietf%3Aparams%3Aoauth%3Agrant-type%3Ajwt-bearer&assertion=" +
                    encodeURI(assertionString),

                    "application/x-www-form-urlencoded" // When accessing the OAuth API, make sure to include this line after the post data. With the Blocks SimpleHTTP, the default mediaType is "application/json".
                )
                // When we get the access_token, we use it to get the calendar events.
                // In this request, we also use timeMinString and timeMaxString to set the time boundaries for what events to load, as well as the Calendar ID set in the spot parameters.
                .then(response => {
                    return SimpleHTTP
                        .newRequest(
                            "https://content.googleapis.com/calendar/v3/calendars/" +
                            encodeURI(spec.spotParameters.calendarId.toString()) +
                            "/events?orderBy=startTime&showDeleted=false&maxResults=100&singleEvents=true&timeMin=" +
                            timeMinString +
                            "&timeMax=" +
                            timeMaxString +
                            "&access_token=" +
                            response.interpreted.access_token, // The access_token obtained from the previous API request

                            { interpretResponse: true }
                        )
                        .setTimeout(25) // Prevents the request from taking to long
                        .get<EventResponse<EventListItem>>()
                        .then(response2 => this.processList(response2.interpreted.items, Math.round(this.timeOffset / 24), Date.parse(timeMaxString))) // Process the list that is returned
                });
	}

	// Mandatory method used to get a single item with full details. In this case, the same request as before is used (after it is processed in the processDetaisl function).
	getDetails(spec: feed.FeedDetailsSpec) {
		return this.processDetails(this.listResponse[spec.index]);
	}

	// Function to process the response we get from the Calendar API, and return a ListItem with just the data we need.
	private processList(list: EventListItem[], currentDay: number, maxTime: any): ListData<ListItem> {
		// Save the API response to a variable, so it can be used again in the processDetails function without making a new, identical, request.
		this.listResponse = list;

		// Create an empty array, which will be filled with info about all events from the day
		const listItems: ListItem[] = [];

		// For every item (event) in the reponse (calendar day), create a new entry in the listItems array with just the data we need, and nothing else.
		for (var item of list) {
			// Format the start- and end times from an ISO-formatted date string to a time string. This will later be formatted further.
			// If no time is available, the event is marked as a full-day event in the Google Calendar. Therefore, we set start to 00:00 and end to 24:00 in that case.
			let startTimeFormatted : string =
				item.start["dateTime"] != undefined ?
					new Date(
						Date.parse(item.start["dateTime"])
					).toTimeString() : "00:00";

			let endTimeFormatted : string =
				item.end["dateTime"] != undefined ?
					new Date(
						Date.parse(item.end["dateTime"])
					).toTimeString() : "24:00";

			// Because timeMin in the API uses the end time, and timeMax uses the start time, we need to make sure that there are no duplicates in the past events and future events
			if (maxTime >= Date.parse(item.end["dateTime"]) || currentDay != 0) {
				// Push to array
				listItems.push(new ListItem(
					item.id,
					item.summary,
					startTimeFormatted.split(":")[0] + ":" + startTimeFormatted.split(":")[1], // Get only the hours and minutes from the time string
					endTimeFormatted.split(":")[0] + ":" + endTimeFormatted.split(":")[1], // Get only the hours and minutes from the time string
				));
			}
		}

		// Return the array of processed items (calendar events)
		return { items: listItems };
	}

	// Similar to processList, but with the htmlLink field added. This processes the same list as processList, but returns it to getDetails instead of getList
	private processDetails(item: DetailsSingleItem): DetailsItem {
		// Format the start- and end times from an ISO-formatted date string to a time string. This will later be formatted further.
		// If no time is available, the event is marked as a full-day event in the Google Calendar. Therefore, we set start to 00:00 and end to 24:00 in that case.
		let startTimeFormatted : string =
			item.start["dateTime"] != undefined ?
				new Date(
					Date.parse(item.start["dateTime"])
				).toTimeString() : "00:00";

		let endTimeFormatted : string =
			item.end["dateTime"] != undefined ?
				new Date(
					Date.parse(item.end["dateTime"])
				).toTimeString() : "24:00";

		// Return the array of processed items (calendar events)
		return {
			id: item.id,
			title: item.summary != undefined ? item.summary : "(No title available)",
			description: item.description,
			startTime: startTimeFormatted.split(":")[0] + ":" + startTimeFormatted.split(":")[1], // Get only the hours and minutes from the time string
			endTime: endTimeFormatted.split(":")[0] + ":" + endTimeFormatted.split(":")[1], // Get only the hours and minutes from the time string
			htmlLink: item.htmlLink.split("?eid=")[1], // The event ID - part of the link in the request response.
			location: item.location != undefined ? item.location : "--"
		};
	}
}

// List of items in the API response
interface EventResponse<T> {
	items: T[];
}

// Class containing a "template" for the items (clendar events), with just the data we need. These fields can be accessed through blocks (unless they are empty)
class ListItem {

	// For each type of data we want to read, create a unique ID visible in the blocks interface.

	// @id("DESCRIPTION VISIBLE IN BLOCKS")
	// readonly ID_HERE_ACCESSABLE_IN_BLOCKS: string;

	@id("Uniquely identifying this feed item")
	readonly id: string;

	@field("Title of the item")
	readonly title: string;

	@field("Event start time")
	readonly startTime: string;

	@field("Event end time")
	readonly endTime: string;

	constructor(id: string, title: string, start: string, end: string) {
		this.id = id;
		this.title = title;
		this.startTime = start;
		this.endTime = end;
	}
}

// What we want out of the getList request
interface EventListItem {
	id: string;
	summary: string;
	description: string;
	htmlLink: string;
	start: string;
	end: string;
}

// Class containing a "template" extending ListItem with more data. These fields can be accessed through blocks (unless they are empty)
class DetailsItem  extends ListItem {
	@field("Event Description")
	readonly description: string;

	@field("Event Id")
	readonly htmlLink: string;

	@field("Location for event")
	readonly location: string;

	constructor(id: string, description: string, title: string, start: string, end: string, htmlLink: string, location: string) {
		super(id, title, start, end),
		this.description = description;
		this.htmlLink = htmlLink;
		this.location = location;
	}
}

// What we want from the getDetails list
interface DetailsSingleItem {
	id: string;
	summary: string;
 	description: string;
 	htmlLink: string;
 	start: string;
 	end: string;
	location: string
}
